#include <assert.h>
#include "perfctr.h"


/*                                                                            */
/*============================================================================*/
/* type definitions                                                           */


typedef struct
  {
    unsigned int MBZ0:32;	       /* 63:32  field with zeroes */
    unsigned int SEL0:1;	       /* 31:31  select 0 */
    unsigned int Ku:1;		       /* 30:30  Kill user counts */
    unsigned int MBZ1:5;	       /* 29:25  field with zeroes */
    unsigned int CBOX2:3;	       /* 24:22  CBOX2 event set if SEL2=0x0F */
    unsigned int CBOX1:3;	       /* 21:19  CBOX1 event set if SEL1=0x0F */
    unsigned int MBZ2:3;	       /* 18:16  field with zeroes */
    unsigned int CTL0:2;	       /* 15:14  control 0 */
    unsigned int CTL1:2;	       /* 13:12  control 1 */
    unsigned int CTL2:2;	       /* 11:10  control 2 */
    unsigned int Kp:1;		       /* 9: 9  Kill PAL counts */
    unsigned int Kk:1;		       /* 8: 8  Kill kernel counts */
    unsigned int SEL1:4;	       /* 7: 4  select 1 */
    unsigned int SEL2:4;	       /* 3: 0  select 2 */
  }
pmctr_in;

typedef struct
  {
    unsigned int CTR0:16;	       /* 63:48  counter 0 */
    unsigned int CTR1:16;	       /* 47:32  counter 1 */
    unsigned int MBZ0:2;	       /* 31:30  field with zeroes */
    unsigned int CTR2:14;	       /* 29:16  counter 2 */
    unsigned int MBZ1:16;	       /* 15: 0  field with zeroes */
  }
pmctr_out;

#define PERFCNT_ON      1
#define PERFCNT_OFF     2
#define PERFCNT_EV5     1

#define CTL_OFF         0
#define CTL_ON          2


/*                                                                            */
/*============================================================================*/
/* start of ev5 performance counters                                          */
/* Returns 0 on success, otherwise some error value                           */


int
perfctr_start(
	       int *sel0,	       /* I: selector counter 0 */
	       int *sel1,	       /* I: selector counter 1 */
	       int *sel2,	       /* I: selector counter 2 */
	       unsigned int mode       /* I: mode bits (kernel,user) */
)

{
  pmctr_in pmctr;
  int ret;
  int mode_field[4][3] = {{0,0,0}, {1,0,0}, {0,1,1}, {0,0,0}};


  /* reset perf counter HW */
  ret = perfmonctl(PERFCNT_EV5, PERFCNT_ON);
  if (ret != 0)
    return ret;

  /* initialize whole control word */
  *(long *) &pmctr = 0;

  /* check mode bits */
  if((mode < 1)  || (mode > 3))
    return 1;

  /* set mode bits */
  pmctr.Kp = mode_field[mode][0];	       /* PAL */
  pmctr.Kk = mode_field[mode][1];;	       /* kernel */
  pmctr.Ku = mode_field[mode][2];	       /* user */

  /* set selector values */
  pmctr.CTL0 = CTL_ON;
  pmctr.SEL0 = *sel0;
  pmctr.CTL1 = CTL_ON;
  if (*sel1 > 0x0e)
    {
      pmctr.SEL1 = 0x0f;
      pmctr.CBOX1 = *sel1 - 0x0f;
    }
  else
    {
      pmctr.SEL1 = *sel1;
      pmctr.CBOX1 = 0;
    }
  pmctr.CTL2 = CTL_ON;
  if (*sel2 > 0x0e)
    {
      pmctr.SEL2 = 0x0f;
      pmctr.CBOX2 = *sel2 - 0x0f;
    }
  else
    {
      pmctr.SEL2 = *sel2;
      pmctr.CBOX2 = 0;
    }


  /* start performance counters */
  return _wrperf(pmctr, 0, 0, 0);
}


/*                                                                            */
/*============================================================================*/
/* reads out performance counters of EV5.                                     */
/* Returns 0 if successful, otherwise some error value.                       */


int
perfctr_stopread(
		  long *ctr0,	       /* O: counter 0 */
		  long *ctr1,	       /* O: counter 1 */
		  long *ctr2	       /* O: counter 2 */
)

{
  pmctr_out pmctr;
  int ret;
  long pc_data[4];


  /* read performance counters */
  ret = _rdperf(pc_data);
  if (ret != 0)
    return ret;

  /* turn performance counters off */
  ret = perfmonctl(PERFCNT_EV5, PERFCNT_OFF);
  if (ret != 0)
    return ret;

  /* NB: at each overflow of a counter, an overflow counter is incremented. */
  /* The overflow counts are returned here in pc_data.  Thus, at       */
  /* completion, the current count of overflows in pc_data[1,2,3]      */
  /* must be added to the current counter values in the pmctr_out      */
  /* structure.                                                        */

  /* retrieve the accumulated overflows and sum their adjusted counts to
     current */
  *(long *) &pmctr = pc_data[0];
  *ctr0 = (pc_data[1] << 16) + pmctr.CTR0;
  *ctr1 = (pc_data[2] << 16) + pmctr.CTR1;
  *ctr2 = (pc_data[3] << 14) + pmctr.CTR2;

  return 0;
}


/*                                                                            */
/*============================================================================*/
