/*==============================================================================


   Purpose:    Performance Counter Library (PCL)


   Author:     Heinz Ziegler, Rudolf Berrendorf
               Research Centre Juelich
               Central Institute for Applied Mathematics
               D-52425 Juelich
               Tel: ++49 2461 61-3218
               r.berrendorf@fz-juelich.de


   last mod:   15.10.1998


   SUPPORT AND DISCLAIMER

   The software in this file (the "Software") is an unsupported product and is
   provided by Zentralinstitut fuer Angewandte Mathematik, Forschungszentrum
   Juelich GmbH (ZAM) "AS IS". We welcome comments, suggestions, and problem
   reports and ask that you mail those to zam@fz-juelich.de; however, we
   make no promises to respond to any item.

   ZAM makes no warrranties or representations, expressed or implied, including,
   but not limited to, implied warranties of merchantability and fitness for a
   particular purpose, as to any element of the Software or any support
   provided in connection with the Software. In no event shall ZAM be
   responsible for any damages, including but not limited to consequential
   damages, arising from or relating to any use of the Software or related
   support.


   COPYRIGHT

   (c) COPYRIGHT 1998
   Zentralinstitut fuer Angewandte Mathematik, Forschungszentrum Juelich GmbH

   The Software is copyrighted. For research purposes, you can use it for free.
   If you plan to use it commercially, please contact zam@fz-juelich.de.


==============================================================================*/

#include "pcl.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>


/*                                                                            */
/*============================================================================*/
/* Include hardware/os dependend libararies, global variables and definitions */

/* define COUNTER_SIZE_32 if the performance counter (in hardware or of the
   low-level driver) has 32 bit counters. Then a timer is activated to avoid
   overflow. The timer interrupts every second to gather performance data.
*/


#if defined(ALPHA_CRAY)
#define PCL_MAX_COUNTER 3
#include <mpp/globals.h>
#include "machines/ALPHA_CRAY/perftest/perfctr.h"


#elif defined(ALPHA_DEC)
#define PCL_MAX_COUNTER 3
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/pfcntr.h>
#include <sys/ioctl.h>
#include <stropts.h>
#include <unistd.h>
#include <sys/processor.h>
#include <sys/resource.h>
#include <sys/sysinfo.h>
#include <machine/hal_sysinfo.h>
static int fd = -1;


#elif defined(MIPS)
#define PCL_MAX_COUNTER 2
int start_counters(int e0, int e1);
int read_counters(int e0, long long *c0, int e1, long long *c1);
static int generation_number;


#elif defined(SPARC)
#define PCL_MAX_COUNTER 2
#include <fcntl.h>
#include <unistd.h>
#include <stropts.h>
#include <sys/types.h>
#include <sys/processor.h>
#include <sys/procset.h>
#include "machines/SPARC/perfmon.h"
/* performance counter device */
static int fd = -1;
/* values of the PIC-registers when performance couting starts */
static unsigned long pic0_value, pic1_value;
#define COUNTER_SIZE_32           /* 32-bit performance counters */


#elif defined(POWERPC)
#define PCL_MAX_COUNTER 4
#include <stdlib.h>
#include "PMencode.h"
#include "PMprocess.h"
#include "PMevents.h"
static pm_int_regs ir;
static pm_event_info_list eil;
static struct PM_PROCESS_INIT pi;
static struct PM_PROCESS_CONTROL pc;
static struct PM_STATUS_CONTROL sc;
/* List of user events */
static pm_event_list uel;
/* List of initial values */
static pm_event_list uivl;
#define COUNTER_SIZE_32           /* 32-bit performance counters */


#elif defined(PENTIUM_LINUX)
#define PCL_MAX_COUNTER 2
#include "machines/PENTIUM_LINUX/libpperf-0.5/libpperf.h"
/* initialized? */
static int initialized = 0;
/* base value of cycle counter */
static long long clock_value;
#define COUNTER_SIZE_32           /* 32-bit performance counters */


#elif defined(PENTIUM_NT)
#define PCL_MAX_COUNTER 2
/* ??? not yet implemented */

#else
#error "unsupported system"

#endif


#if defined(COUNTER_SIZE_32)
#include <signal.h>
#endif



/*                                                                            */
/*============================================================================*/
/* Global system independent variables                                        */


/* PCL version number */
char *PCL_version_number = "1.0";

/* count mode (default: user mode) */
unsigned int count_mode = PCL_MODE_USER;

/* nesting level of PCL routines. Is incremented when PCL_start is called,
   and decremented when PCL_stop is called
 */
static int nesting_level = -1;

/* number of requested events at each nesting level */
static int event_count;

/* requested events at each nesting level */
static int events[PCL_MAX_EVENT_PER_CALL];

/* partial result values when stop/starting */
static PCL_CNT_TYPE partial_results[PCL_MAX_NESTING_LEVEL][PCL_MAX_EVENT_PER_CALL];

#if defined(COUNTER_SIZE_32)           /* 32-bit performance counters */
/* old signal function */
void (*old_signal_handler) (int);
#endif


/*----------------------------------------------------------------------------*/
/* For each event, there exists a table entry with the following components:.
   If the event is in some way available on the computer, "available" is
   set to 1.
   If the event is mapped directly to a low-level counter, "composed" is
   set to 0 and "data.command" contains the appropriate command word for the
   driver. Each entry not equal to -1 is a possible command word. As one event
   may be counted on different low-level counters, several command[i] may be
   different to -1.
   If the event is measured by a combination of some low-level counters,
   "composed" is set to 1 and the rest of the data is stored in "composed_data".
   "composed_data.command" contains the command words and
   "composed.data.operation contains the connection operation of the
   appropriate low-level counter values (e.g. (+1) * ctr_0 + (-1) * ctr_1).
   Example for non-composed entry:
   {PCL_AVAIL, PCL_INTEG, {0x02, 0x03}}
   The event may be counted on counter 0 or counter 1, the command words for
   the low-level driver are 0x02 (for counter 0) and 0x03 (counter 1),
   respectively.
   Example for composed entry:
   {PCL_AVAIL, PCL_COMP, {{0x0e, 0x0f}, {1, -1}}}
   The event is composed out of two events. The command words for
   the low-level driver are 0x0e (for counter 0) and 0x0f (counter 1),
   respectively. The resulting counter values (ctr_0, ctr_1) are connected in
   the form: (+1) * ctr_0 + (-1) * ctr_1
 */

typedef struct
  {
    int available;		          /* event is available in HW/driver? */
    int composed;		          /* event is composed? */
    union
      {
	struct
	  {
	    long command[PCL_MAX_COUNTER];    /* command values */
	    int operation[PCL_MAX_COUNTER];	  /* operations (+1,-1) */
	  }
	composed_data;
	long command[PCL_MAX_COUNTER];        /* command value */
      }
    data;
  }
event_descr_t;


/* table of event entries (see above):
static event_descr_t event_table[PCL_MAX_EVENT];
*/

/*----------------------------------------------------------------------------*/
/* Inverse table entries. Given the results in ctr_0,...,ctr_n, this table
   specifies, how the PCL result values are constructed and where they are
   stored. Each hardware counter has such an entry.
 */

typedef struct
  {
    int used;			       /* 1: value is used */
    long command;                      /* command to activate low-level cntr */
    PCL_CNT_TYPE result_value;	       /* result value of low-level driver */
    int result_index;		       /* contributes to what user result */
    int operation;		       /* +1, -1 */
    int event;                         /* used on what event */
  }
result_descr_t;


/* for each low-level counter one entry */
static result_descr_t results[PCL_MAX_COUNTER];


/*----------------------------------------------------------------------------*/

/* marker in table if this event is not available/available */
#define PCL_UNAVAIL      0
#define PCL_AVAIL        1


/*----------------------------------------------------------------------------*/

#if defined(ALPHA_CRAY)

/* allowed count modes */
unsigned int allowed_count_mode = PCL_MODE_USER_SYSTEM;

/* invalid entry-data if event is not available */
#define PCL_INVALID    0, {{{-1,-1,-1},{0,0,0}}}

/* ??? check table */
/* entries are control codes for hardware performance counter.
   sel0-sel2 contain event selectors for counter0 - counter2. For counter1
   and counter2, CBOX events are decoded as event selectors larger than 0x0e.
   The CBOX value is then computed as (value - 0x0f).
*/
static event_descr_t event_table[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L1CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L1CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L1CACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L1DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L1DCACHE_WRITE */
  {PCL_AVAIL,0,{{{-1,0xE,-1},{-1,1,-1}}}},       /* PCL_L1DCACHE_READWRITE */
  {PCL_AVAIL,1,{{{-1,0xE,0x6},{-1,1,-1}}}},      /* PCL_L1DCACHE_HIT */
  {PCL_AVAIL,0,{{{-1,-1,0x6},{-1,-1,1}}}},       /* PCL_L1DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L1DCACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L1ICACHE_WRITE */
  {PCL_AVAIL,0,{{{-1,0xD,-1},{-1,1,-1}}}},       /* PCL_L1ICACHE_READWRITE */
  {PCL_AVAIL,1,{{{-1,0xD,0x3},{-1,1,-1}}}},      /* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,0,{{{-1,-1,0x3},{-1,-1,1}}}},       /* PCL_L1ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L1ICACHE_USED_LINES */
  {PCL_AVAIL,0,{{{-1,0x10,-1},{-1,1,-1}}}},      /* PCL_L2CACHE_READ */
  {PCL_AVAIL,0,{{{-1,0x11,-1},{-1,1,-1}}}},      /* PCL_L2CACHE_WRITE */
  {PCL_AVAIL,0,{{{-1,0xF,-1},{-1,1,-1}}}},       /* PCL_L2CACHE_READWRITE */
  {PCL_AVAIL,1,{{{-1,0xF,0xF},{-1,1,-1}}}},      /* PCL_L2CACHE_HIT */
  {PCL_AVAIL,0,{{{-1,-1,0xF},{-1,-1,1}}}},       /* PCL_L2CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2CACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2DCACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_L2ICACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_TLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_ITLB_HIT */
  {PCL_AVAIL,0,{{{-1,-1,0x5},{-1,-1,1}}}},       /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_DTLB_HIT */
  {PCL_AVAIL,0,{{{-1,-1,0x7},{-1,-1,1}}}},       /* PCL_DTLB_MISS */
  {PCL_AVAIL,0,{{{0x0,-1,-1},{1,-1,-1}}}},       /* PCL_CYCLES */
  {PCL_AVAIL,0,{{{-1,0x9,-1},{-1,1,-1}}}},       /* PCL_INTEGER_INSTR */
  {PCL_AVAIL,0,{{{-1,0xA,-1},{-1,1,-1}}}},       /* PCL_FP_INSTR */
  {PCL_AVAIL,0,{{{-1,0xB,-1},{-1,1,-1}}}},       /* PCL_LOAD_INSTR */
  {PCL_AVAIL,0,{{{-1,0xC,-1},{-1,1,-1}}}},       /* PCL_STORE_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,0,{{{0x1,-1,-1},{1,-1,-1}}}},       /* PCL_INSTR */
  {PCL_AVAIL, PCL_INVALID},                      /* PCL_JUMP_SUCCESS */
  {PCL_AVAIL,0,{{{-1,-1,0x2},{-1,-1,1}}}},       /* PCL_JUMP_UNSUCCESS */
  {PCL_AVAIL, PCL_INVALID},                      /* PCL_JUMP */
  {PCL_AVAIL,0,{{{-1,-1,0xE},{-1,-1,1}}}},       /* PCL_ATOMIC_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_ATOMIC_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_STALL_INTEGER */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_STALL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_STALL_LOAD */
  {PCL_UNAVAIL, PCL_INVALID},                    /* PCL_STALL_STORE */
  {PCL_UNAVAIL, PCL_INVALID}                     /* PCL_STALL */
};

/*----------------------------------------------------------------------------*/

#elif defined(ALPHA_DEC)

/* allowed count modes */
unsigned int allowed_count_mode =  PCL_MODE_USER_SYSTEM;

/* invalid entry-data if event is not available */
#define PCL_INVALID    0, {{{-1,-1,-1},{0,0,0}}}

/* entries are control codes for hardware performance counter */
static event_descr_t event_table[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1CACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1DCACHE_WRITE */
  {PCL_AVAIL,0,{{{-1,PF5_MUX1_C_DCACHE_ACCESS,-1},{-1,1,-1}}}},	/* PCL_L1DCACHE_READWRITE */
  {PCL_AVAIL,1,{{{-1,PF5_MUX1_C_DCACHE_ACCESS,PF5_MUX2_C_DCACHE_MISSES},{-1,1,-1}}}}, /* PCL_L1DCACHE_HIT */
  {PCL_AVAIL,0,{{{-1,-1,PF5_MUX2_C_DCACHE_MISSES},{-1,-1,1}}}},	/* PCL_L1DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1DCACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1ICACHE_WRITE */
  {PCL_AVAIL,0,{{{-1,PF5_MUX1_C_ICACHE_ACCESS,-1},{-1,1,-1}}}},	/* PCL_L1ICACHE_READWRITE */
  {PCL_AVAIL,1,{{{-1,PF5_MUX1_C_ICACHE_ACCESS,PF5_MUX2_C_ICACHE_MISSES},{-1,1,-1}}}}, /* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,0,{{{-1,-1,PF5_MUX2_C_ICACHE_MISSES},{-1,-1,1}}}},	/* PCL_L1ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L1ICACHE_USED_LINES */
  {PCL_AVAIL,0,{{{-1,PF5_MUX1_C_SCACHE_READ,-1},{-1,1,-1}}}},	/* PCL_L2CACHE_READ */
  {PCL_AVAIL,0,{{{-1,PF5_MUX1_C_SCACHE_WRITE,-1},{-1,1,-1}}}},	/* PCL_L2CACHE_WRITE */
  {PCL_AVAIL,0,{{{-1,PF5_MUX1_C_SCACHE_ACCESS,-1},{-1,1,-1}}}},	/* PCL_L2CACHE_READWRITE */
  {PCL_AVAIL,0,{{{-1,PF5_MUX1_C_SCACHE_ACCESS,PF5_MUX2_C_SCACHE_MISS},{-1,1,-1}}}}, /* PCL_L2CACHE_HIT */
  {PCL_AVAIL,0,{{{-1,-1,PF5_MUX2_C_SCACHE_MISS},{-1,-1,1}}}},	/* PCL_L2CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2CACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2DCACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_L2ICACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_TLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_ITLB_HIT */
  {PCL_AVAIL,0,{{{-1,-1,PF5_MUX2_C_ITB_MISS},{-1,-1,1}}}}, /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_DTLB_HIT */
  {PCL_AVAIL,0,{{{-1,-1,PF5_MUX2_C_DTB_MISS},{-1,-1,1}}}},	/* PCL_DTLB_MISS */
  {PCL_AVAIL,0,{{{PF5_MUX0_CYCLES,-1,-1},{1,-1,-1}}}},	/* PCL_CYCLES */
  {PCL_AVAIL,0,{{{-1,PF5_MUX1_INTEGER_OPERATE,-1},{-1,1,-1}}}},	/* PCL_INTEGER_INSTR */
  {PCL_AVAIL,0,{{{-1,PF5_MUX1_C_FP_INSTRUCTIONS,-1},{-1,1,-1}}}}, /* PCL_FP_INSTR */
  {PCL_AVAIL,0,{{{-1,PF5_MUX1_C_LOAD_INSTRUCTIONS,-1},{-1,1,-1}}}}, /* PCL_LOAD_INSTR */
  {PCL_AVAIL,0,{{{-1,PF5_MUX1_C_STORE_INSTRUCTIONS,-1},{-1,1,-1}}}}, /* PCL_STORE_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,0,{{{PF5_MUX0_ISSUES,-1,-1},{1,-1,-1}}}}, /* PCL_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_JUMP_SUCCESS */
  {PCL_AVAIL,0,{{{-1,-1,PF5_MUX2_C_BRANCH_MISPREDICTS},{-1,-1,1}}}}, /* PCL_JUMP_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_JUMP */
  {PCL_AVAIL,0,{{{-1,-1,PF5_MUX2_C_LOAD_LOCKED},{-1,-1,1}}}},	/* PCL_ATOMIC_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_ATOMIC_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_STALL_INTEGER */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_STALL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_STALL_LOAD */
  {PCL_UNAVAIL, PCL_INVALID},                     /* PCL_STALL_STORE */
  {PCL_UNAVAIL, PCL_INVALID}                      /* PCL_STALL */
};

/*----------------------------------------------------------------------------*/

#elif defined(MIPS)

/* allowed count modes */
unsigned int allowed_count_mode = PCL_MODE_USER;

/* invalid entry-data if event is not available */
#define PCL_INVALID    0, {{{-1,-1},{0,0}}}

/* entries are control codes for hardware performance counter */
static event_descr_t event_table[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_HIT */
  {PCL_AVAIL,1,{{{9,25},{1,1}}}},      /* PCL_L1CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_HIT */
  {PCL_AVAIL,0,{{{-1,25},{-1,1}}}},    /* PCL_L1DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,0,{{{9,-1},{1,-1}}}},     /* PCL_L1ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_HIT */
  {PCL_AVAIL,1,{{{10,26},{1,1}}}},     /* PCL_L2CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_HIT */
  {PCL_AVAIL,0,{{{-1,26},{-1,1}}}},    /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_HIT */
  {PCL_AVAIL,0,{{{10,-1},{1,-1}}}},    /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_TLB_HIT */
  {PCL_AVAIL,0,{{{23,-1},{1,-1}}}},    /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ITLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_DTLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_DTLB_MISS */
  {PCL_AVAIL,0,{{{0,0},{1,1}}}},       /* PCL_CYCLES */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_INTEGER_INSTR */
  {PCL_AVAIL,0,{{{-1,21},{-1,1}}}},    /* PCL_FP_INSTR */
  {PCL_AVAIL,0,{{{-1,18},{-1,1}}}},    /* PCL_LOAD_INSTR */
  {PCL_AVAIL,0,{{{-1,19},{-1,1}}}},    /* PCL_STORE_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,0,{{{-1,17},{-1,1}}}},    /* PCL_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_JUMP_SUCCESS */
  {PCL_AVAIL,0,{{{-1,24},{-1,1}}}},    /* PCL_JUMP_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},          /* PCL_JUMP */
  {PCL_AVAIL,0,{{{-1,20},{-1,1}}}},    /* PCL_ATOMIC_SUCCESS */
  {PCL_AVAIL,0,{{{5,-1},{1,-1}}}},     /* PCL_ATOMIC_UNSUCCESS */
  {PCL_AVAIL,1,{{{5,20},{1,1}}}},      /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_INTEGER */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_LOAD */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_STORE */
  {PCL_UNAVAIL, PCL_INVALID}	       /* PCL_STALL */
};

/*----------------------------------------------------------------------------*/

#elif defined(SPARC)

/* allowed count modes */
unsigned int allowed_count_mode = PCL_MODE_USER_SYSTEM;

/* invalid entry-data if event is not available */
#define PCL_INVALID    0, {{{-1, -1},{0,0}}}

static event_descr_t event_table[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1CACHE_USED_LINES */
  {PCL_AVAIL,0,{{{PCR_S0_DC_READ,-1},{1,-1}}}},    /* PCL_L1DCACHE_READ */
  {PCL_AVAIL,0,{{{PCR_S0_DC_WRITE,-1},{1,-1}}}},   /* PCL_L1DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1DCACHE_HIT */
  {PCL_AVAIL,0,{{{PCR_S0_EC_READ_HIT,-1},{1,-1}}}}, /* PCL_L1DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1DCACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1ICACHE_WRITE */
  {PCL_AVAIL,0,{{{PCR_S0_IC_REF,-1},{1,-1}}}},     /* PCL_L1ICACHE_READWRITE */
  {PCL_AVAIL,0,{{{-1,PCR_S1_IC_HIT},{-1,1}}}},     /* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,1,{{{PCR_S0_IC_REF,PCR_S1_IC_HIT},{1,-1}}}}, /* PCL_L1ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L1ICACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2CACHE_WRITE */
  {PCL_AVAIL,0,{{{PCR_S0_EC_REF,-1},{1,-1}}}},     /* PCL_L2CACHE_READWRITE */
  {PCL_AVAIL,0,{{{-1,PCR_S1_EC_HIT},{-1,1}}}},     /* PCL_L2CACHE_HIT */
  {PCL_AVAIL,0,{{{-1,PCR_S1_EC_WRITEBACK},{-1,1}}}}, /* PCL_L2CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2CACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2DCACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_L2ICACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_TLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_ITLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_DTLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_DTLB_MISS */
  {PCL_AVAIL,0,{{{PCR_S0_CYCLE_CNT,PCR_S1_CYCLE_CNT},{1,1}}}}, /* PCL_CYCLES */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_INTEGER_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_FP_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_LOAD_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_STORE_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,0,{{{PCR_S0_INSTR_CNT,-1},{1,-1}}}},  /* PCL_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_JUMP_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_JUMP_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_ATOMIC_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_ATOMIC_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_STALL_INTEGER */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_STALL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_STALL_LOAD */
  {PCL_UNAVAIL, PCL_INVALID},	                   /* PCL_STALL_STORE */
  {PCL_UNAVAIL, PCL_INVALID}	                   /* PCL_STALL */
};


/*----------------------------------------------------------------------------*/

#elif defined(POWERPC)

/* allowed count modes */
unsigned int allowed_count_mode = PCL_MODE_USER_SYSTEM;

/* bringing PCL and PMapi together, you need x different events (countable on
   counter x only) to actually count nothing on x counters
*/
static long default_event[PCL_MAX_COUNTER] =
 {
   PM_IC_MISS,
   PM_DC_MISS,
   PM_FPU_WT,
   PM_BTAC_HITS
 };

/* invalid entry-data if event is not available */
#define PCL_INVALID    0, {{{-1, -1, -1, -1},{0,0,0,0}}}

static event_descr_t event_table[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1CACHE_HIT */
  {PCL_AVAIL,1,{{{PM_IC_MISS,PM_DC_MISS,-1,-1},{1,1,-1,-1}}}}, /* PCL_L1CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1CACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1DCACHE_HIT */
  {PCL_AVAIL,0,{{{-1,PM_DC_MISS,-1,-1},{-1,1,-1,-1}}}}, /* PCL_L1DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1DCACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,0,{{{PM_IC_MISS,-1,-1,-1},{1,-1,-1,-1}}}}, /* PCL_L1ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L1ICACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2CACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2DCACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_L2ICACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_TLB_HIT */
  {PCL_AVAIL,1,{{{PM_DTLB_MISS,PM_ITLB_MISS,-1,-1},{1,1,-1,-1}}}}, /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_ITLB_HIT */
  {PCL_AVAIL,0,{{{-1,PM_ITLB_MISS,-1,-1},{-1,1,-1,-1}}}}, /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_DTLB_HIT */
  {PCL_AVAIL,0,{{{PM_DTLB_MISS,-1,-1,-1},{1,-1,-1,-1}}}}, /* PCL_DTLB_MISS */
  {PCL_AVAIL,0,{{{PM_CYC,PM_CYC,PM_CYC,PM_CYC},{1,1,1,1}}}}, /* PCL_CYCLES */
  {PCL_AVAIL,0,{{{PM_FXU_CMPL,-1,-1,-1},{1,-1,-1,-1}}}}, /* PCL_INTEGER_INSTR */
  {PCL_AVAIL,0,{{{PM_FPU_CMPL,-1,-1,-1},{1,-1,-1,-1}}}}, /* PCL_FP_INSTR */
  {PCL_AVAIL,0,{{{PM_LS_EXEC,-1,-1,-1},{1,-1,-1,-1}}}}, /* PCL_LOAD_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_STORE_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,0,{{{PM_INST_DISP,-1,-1,-1},{1,-1,-1,-1}}}}, /* PCL_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_JUMP_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_JUMP_UNSUCCESS */
  {PCL_AVAIL,0,{{{-1,PM_BR_CMPL,-1,-1},{-1,1,-1,-1}}}}, /* PCL_JUMP */
  {PCL_AVAIL,0,{{{-1,PM_RESRV_CMPL,-1,-1},{-1,1,-1,-1}}}}, /* PCL_ATOMIC_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_ATOMIC_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_STALL_INTEGER */
  {PCL_AVAIL,0,{{{-1,-1,PM_FPU_WT,-1},{-1,-1,1,-1}}}}, /* PCL_STALL_FP */
  {PCL_AVAIL,0,{{{-1,-1,PM_BRU_WT,-1},{-1,-1,1,-1}}}}, /* PCL_STALL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_STALL_LOAD */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_STALL_STORE */
  {PCL_UNAVAIL, PCL_INVALID}                 /* PCL_STALL */
};

/*----------------------------------------------------------------------------*/

#elif defined(PENTIUM_LINUX)

/* allowed count modes */
unsigned int allowed_count_mode = PCL_MODE_USER_SYSTEM;

/* entries are event names for libpperf */
/* Distinguish at runtime between Pentium [MMX] and Pentium Pro/Pentium II
   CPUs 
 */
static event_descr_t *event_table = NULL;

/* invalid entry-data if event is not available */
#define PCL_INVALID    0, {{{-1, -1},{0,0}}}

static event_descr_t event_table_pentium[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_USED_LINES */
  {PCL_AVAIL,0,{{{P5_MEM_DATA_READ,P5_MEM_DATA_READ},{1,1}}}}, /* PCL_L1DCACHE_READ */
  {PCL_AVAIL,0,{{{P5_MEM_DATA_WRITE,P5_MEM_DATA_WRITE},{1,1}}}}, /* PCL_L1DCACHE_WRITE */
  {PCL_AVAIL,1,{{{P5_MEM_DATA_READ,P5_MEM_DATA_WRITE},{1,1}}}}, /* PCL_L1DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L1DCACHE_HIT */
  {PCL_AVAIL,0,{{{P5_MEM_READ_WRITE_MISS,P5_MEM_READ_WRITE_MISS},{1,1}}}}, /* PCL_L1DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L1DCACHE_USED_LINES */
  {PCL_AVAIL,0,{{{P5_CODE_READ,P5_CODE_READ},{1,1}}}}, /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L1ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L1ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,0,{{{P5_CODE_CACHE_MISS,P5_CODE_CACHE_MISS},{1,1}}}}, /* PCL_L1ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L1ICACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L2CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L2CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L2CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L2CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L2CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L2CACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L2DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L2DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L2DCACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L2ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L2ICACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_TLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_ITLB_HIT */
  {PCL_AVAIL,0,{{{P5_CODE_TLB_MISS,P5_CODE_TLB_MISS},{1,1}}}}, /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_DTLB_HIT */
  {PCL_AVAIL,0,{{{P5_TLB_MISS,P5_TLB_MISS},{1,1}}}}, /* PCL_DTLB_MISS */
  /* special case with dummy entry: handled in stop_counting */
  {PCL_AVAIL,0,{{{P5_TLB_MISS,P5_TLB_MISS},{1,1}}}}, /* PCL_CYCLES */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_INTEGER_INSTR */
  {PCL_AVAIL,0,{{{P5_FLOPS,P5_FLOPS},{1,1}}}}, /* PCL_FP_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_LOAD_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_STORE_INSTR */
  {PCL_AVAIL,0,{{{P5_MEM_READ_WRITE_HIT,P5_MEM_READ_WRITE_HIT},{1,1}}}}, /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,0,{{{P5_INS_EXE_B_PIPES,P5_INS_EXE_B_PIPES},{1,1}}}}, /* PCL_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_JUMP_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},                /* PCL_JUMP_UNSUCCESS */
  {PCL_AVAIL,0,{{{P5_BRANCHES,P5_BRANCHES},{1,1}}}}, /* PCL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_ATOMIC_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_ATOMIC_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_STALL_INTEGER */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_STALL_JUMP */
  {PCL_AVAIL,0,{{{P5_PIPE_STL_WDR,P5_PIPE_STL_WDR},{1,1}}}}, /* PCL_STALL_LOAD */
  {PCL_AVAIL,0,{{{P5_PIPE_STL_FWB,P5_PIPE_STL_FWB},{1,1}}}}, /* PCL_STALL_STORE */
  {PCL_UNAVAIL, PCL_INVALID}	             /* PCL_STALL */
};

static event_descr_t event_table_pentium2[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L1CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L1CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L1CACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L1DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L1DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L1DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L1DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L1DCACHE_MISS */
  {PCL_AVAIL,0,{{{P6_DCU_LINES_IN,P6_DCU_LINES_IN},{1,1}}}}, /* PCL_L1DCACHE_USED_LINES */
  {PCL_AVAIL,0,{{{P6_IFU_IFETCH,P6_IFU_IFETCH},{1,1}}}}, /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L1ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L1ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,0,{{{P6_IFU_IFETCH_MISS,P6_IFU_IFETCH_MISS},{1,1}}}}, /* PCL_L1ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L1ICACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L2CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L2CACHE_WRITE */
  {PCL_AVAIL,0,{{{P6_L2_RQSTS,P6_L2_RQSTS},{1,1}}}}, /* PCL_L2CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L2CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L2CACHE_MISS */
  {PCL_AVAIL,0,{{{P6_L2_LINES_IN,P6_L2_LINES_IN},{1,1}}}}, /* PCL_L2CACHE_USED_LINES */
  {PCL_AVAIL,0,{{{P6_L2_LD,P6_L2_LD},{1,1}}}}, /* PCL_L2DCACHE_READ */
  {PCL_AVAIL,0,{{{P6_L2_ST,P6_L2_ST},{1,1}}}}, /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L2DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L2DCACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L2ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_L2ICACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_TLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_ITLB_HIT */
  {PCL_AVAIL,0,{{{P6_ITLB_MISS,P6_ITLB_MISS},{1,1}}}}, /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_DTLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_DTLB_MISS */
  /* special case with dummy entry: handled in stop_counting */
  {PCL_AVAIL,0,{{{P6_CPU_CLK_UNHALTED,P6_CPU_CLK_UNHALTED},{1,1}}}}, /* PCL_CYCLES */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_INTEGER_INSTR */
  {PCL_AVAIL,0,{{{P6_COMP_FLOP_RET,P6_COMP_FLOP_RET},{1,1}}}}, /* PCL_FP_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_LOAD_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_STORE_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,0,{{{P6_INST_RETIRED,P6_INST_RETIRED},{1,1}}}}, /* PCL_INSTR */
  {PCL_AVAIL,0,{{{P6_BR_TAKEN_RETIRED,P6_BR_TAKEN_RETIRED},{1,1}}}}, /* PCL_JUMP_SUCCESS */
  {PCL_AVAIL, 0,{{{P6_BR_MISS_PRED_TAKEN_RET,P6_BR_MISS_PRED_TAKEN_RET},{1,1}}}}, /* PCL_JUMP_UNSUCCESS */
  {PCL_AVAIL,1,{{{P6_BR_TAKEN_RETIRED,P6_BR_MISS_PRED_TAKEN_RET},{1,1}}}}, /* PCL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_ATOMIC_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_ATOMIC_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_STALL_INTEGER */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_STALL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_STALL_LOAD */
  {PCL_UNAVAIL, PCL_INVALID},	             /* PCL_STALL_STORE */
  {PCL_AVAIL,0,{{{P6_RESOURCE_STALLS,P6_RESOURCE_STALLS},{1,1}}}} /* PCL_STALL */
};

/*----------------------------------------------------------------------------*/

#elif defined(PENTIUM_NT)

/* allowed count modes */
unsigned int allowed_count_mode = PCL_MODE_USER_SYSTEM;

/* entries are control codes for hardware performance counter */
/* Distinguish at runtime between Pentium [MMX] and Pentium Pro/Pentium II
   CPUs 
 */
static event_descr_t *event_table = NULL;

/* invalid entry-data if event is not available */
#define PCL_INVALID    0, {{{-1, -1},{0,0}}}

static event_descr_t event_table_pentium[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_USED_LINES */
  {PCL_AVAIL,0,{{{0x00,0x00},{1,1}}}}, /* PCL_L1DCACHE_READ */
  {PCL_AVAIL,0,{{{0x01,0x01},{1,1}}}}, /* PCL_L1DCACHE_WRITE */
  {PCL_AVAIL,1,{{{0x00, 0x01},{1,1}}}}, /* PCL_L1DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_HIT */
  {PCL_AVAIL,0,{{{0x29,0x29},{1,1}}}}, /* PCL_L1DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_USED_LINES */
  {PCL_AVAIL,0,{{{0x0C,0x0C},{1,1}}}}, /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,0,{{{0x0E,0x0E},{1,1}}}}, /* PCL_L1ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_TLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ITLB_HIT */
  {PCL_AVAIL,0,{{{0x0D,0x0D},{1,1}}}}, /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_DTLB_HIT */
  {PCL_AVAIL,0,{{{0x02,0x02},{1,1}}}}, /* PCL_DTLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_CYCLES */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_INTEGER_INSTR */
  {PCL_AVAIL,0,{{{0x22,0x22},{1,1}}}}, /* PCL_FP_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_LOAD_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STORE_INSTR */
  {PCL_AVAIL,0,{{{0x28,0x28},{1,1}}}}, /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,0,{{{0x16,0x16},{1,1}}}}, /* PCL_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_JUMP_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_JUMP_UNSUCCESS */
  {PCL_AVAIL,0,{{{0x12,0x12},{1,1}}}}, /* PCL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_INTEGER */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_JUMP */
  {PCL_AVAIL,0,{{{0x1A,0x1A},{1,1}}}}, /* PCL_STALL_LOAD */
  {PCL_AVAIL,0,{{{0x19,0x19},{1,1}}}}, /* PCL_STALL_STORE */
  {PCL_UNAVAIL, PCL_INVALID}	       /* PCL_STALL */
};

static event_descr_t event_table_pentium2[PCL_MAX_EVENT] =
{
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1CACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1DCACHE_MISS */
  {PCL_AVAIL,0,{{{0x45,0x45},{1,1}}}}, /* PCL_L1DCACHE_USED_LINES */
  {PCL_AVAIL,0,{{{0x80,0x80},{1,1}}}}, /* PCL_L1ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_HIT */
  {PCL_AVAIL,0,{{{0x81,0x81},{1,1}}}}, /* PCL_L1ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L1ICACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_WRITE */
  {PCL_AVAIL,0,{{{0x2E,0x2E},{1,1}}}}, /* PCL_L2CACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2CACHE_MISS */
  {PCL_AVAIL,0,{{{0x24,0x24},{1,1}}}}, /* PCL_L2CACHE_USED_LINES */
  {PCL_AVAIL,0,{{{0x29,0x29},{1,1}}}}, /* PCL_L2DCACHE_READ */
  {PCL_AVAIL,0,{{{0x2A,0x2A},{1,1}}}}, /* PCL_L2DCACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2DCACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_READ */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_WRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_READWRITE */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_L2ICACHE_USED_LINES */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_TLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_TLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ITLB_HIT */
  {PCL_AVAIL,0,{{{0x85,0x85},{1,1}}}}, /* PCL_ITLB_MISS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_DTLB_HIT */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_DTLB_MISS */
  {PCL_AVAIL,0,{{{0x79,0x79},{1,1}}}}, /* PCL_CYCLES */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_INTEGER_INSTR */
  {PCL_AVAIL,0,{{{0xC1,-1},{1,-1}}}},  /* PCL_FP_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_LOAD_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STORE_INSTR */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_LOADSTORE_INSTR */
  {PCL_AVAIL,0,{{{0xC0,0xC0},{1,1}}}}, /* PCL_INSTR */
  {PCL_AVAIL,0,{{{0xC9,0xC9},{1,1}}}}, /* PCL_JUMP_SUCCESS */
  {PCL_AVAIL,0,{{{0xCA,0xCA},{1,1}}}}, /* PCL_JUMP_UNSUCCESS */
  {PCL_AVAIL,0,{{{0xC9,0xCA},{1,1}}}}, /* PCL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC_SUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC_UNSUCCESS */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_ATOMIC */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_INTEGER */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_FP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_JUMP */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_LOAD */
  {PCL_UNAVAIL, PCL_INVALID},	       /* PCL_STALL_STORE */
  {PCL_AVAIL,0,{{{0xA2,0xA2},{1,1}}}}  /* PCL_STALL */
};

/*----------------------------------------------------------------------------*/

#endif


/*                                                                            */
/*============================================================================*/
/* forward declarations */


static int start_counting(void);
static int stop_counting(
			  PCL_CNT_TYPE * results_list,	/* O: counter values */
			  int ncounter /* I: number of event counters */
);


/*                                                                            */
/*============================================================================*/
/* Determine which type of Pentium processor we are running.                  */
/* Sets "event_table" accordingly to get the right register settings.         */
/* Returns 0 if processor type is supported.                                  */


#if defined(PENTIUM_LINUX)

static int
determine_pentium_type(void)

{
  unsigned int cpu;


  /* determine CPU type */
  asm("cpuid\n\t"
      : "=a"(cpu)
      : "a"(0)
      : "%eax", "%ebx", "%ecx", "%edx");

  if (cpu == 1)
    /* Pentium */
    event_table = event_table_pentium;

  else if (cpu >= 2)
    /* Pentium Pro, Pentium II */
    event_table = event_table_pentium2;

  else
    /* something different */
    return 1;

  return 0;
}

#endif


/*                                                                            */
/*============================================================================*/
/* interrupt routine to avoid overflow.                                       */


#if defined(COUNTER_SIZE_32)           /* 32-bit performance counters */

static void
interrupt_handling(
		    int sig
)

{
  int i, ncounter;
  PCL_CNT_TYPE tmp_results_list[PCL_MAX_EVENT_PER_CALL];


  /* get number of counters on this nesting level */
  ncounter = event_count;

  /* stop counting and read values */
  if (stop_counting(tmp_results_list, ncounter) != PCL_SUCCESS)
    fprintf(stderr, "PCL: problem in interrupt routine (1)\n");

  /* add to partial results */
  for (i = 0; i < ncounter; ++i)
    partial_results[nesting_level][i] += tmp_results_list[i];

  /* restart counting */
  if (start_counting() != PCL_SUCCESS)
    fprintf(stderr, "PCL: problem in interrupt routine (2)\n");
}

#endif


/*                                                                            */
/*============================================================================*/
/* Try to match the event requests with the available counters.               */
/* First, try to match composed events as these are fixed to certain counters.*/
/* Then, try to match integral events to left counters.                       */


static int
get_low_level_counters(
			int *counter_list,	/* I: events to be started */
			int ncounter,   /* I: number of events */
			result_descr_t results[PCL_MAX_COUNTER] /* O: cntr descr */
)

{
  int i, j, ecount, found;
  event_descr_t *event;


  ecount = 0;

  /* initialize result descriptors */
  for (i = 0; i < PCL_MAX_COUNTER; ++i)
    results[i].used = 0;


  /*--------------------------------------------------------------------------*/
  /* handle each composed request */

  for (i = 0; i < ncounter; ++i)
    {
      /* get event descriptor entry for i-th event request */
      event = &event_table[counter_list[i]];

      /* check if event is supported */
      if ((event->available == PCL_AVAIL) && (event->composed == 1))
	{
	  /* check for low level counters which match event */
	  for (j = 0; j < PCL_MAX_COUNTER; ++j)
	    {
	      if (event->data.composed_data.command[j] != -1)
		{
		  if (results[j].used == 0)
		    {
		      /* counter found */
		      /* this counter will be  used */
		      results[j].used = 1;

		      /* command to activate cunter */
		      results[j].command = event->data.composed_data.command[j];

		      /* counter value is used for result i */
		      results[j].result_index = i;

		      /* operation (+1, -1) */
		      results[j].operation =
			event->data.composed_data.operation[j];

		      /* store to what event type it belongs */
		      results[j].event = counter_list[i];
		    }

		  else
		    /* counter already used */
		    return PCL_NOT_SUPPORTED;
		}		       /* if(event->data.composed... */
	    }			       /* for(j = 0; j < PCL_MAX_COUNTER; ++j) */

	  /* one more event done */
	  ++ecount;
	}
    }



  /*--------------------------------------------------------------------------*/
  /* handle each integral request */

  for (i = 0; i < ncounter; ++i)
    {
      /* get event descriptor entry for i-th event request */
      event = &event_table[counter_list[i]];

      /* check if event is supported */
      if ((event->available == PCL_AVAIL) && (event->composed == 0))
	{
	  /* check for low level counter which matches event and  which is
	     unused */
	  found = 0;
	  for (j = 0; (j < PCL_MAX_COUNTER) && !found; ++j)
	    {
	      if ((event->data.command[j] != -1) && (results[j].used == 0))
		{
		  /* counter found */
		  /* this counter will be  used */
		  results[j].used = 1;

		  /* command to activate cunter */
		  results[j].command = event->data.command[j];

		  /* counter value is used for result i */
		  results[j].result_index = i;

		  /* integral value ((+1)*value)  */
		  results[j].operation = +1;

		  /* store to what event type it belongs */
		  results[j].event = counter_list[i];

		  /* one more event done */
		  ++ecount;

		  /* leave loop */
		  found = 1;
		}
	    }
	}
    }


  if (ecount != ncounter)
    /* at least one event is not supported */
    return PCL_NOT_SUPPORTED;


  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* Based on the result values of the low-level counters, calculate results.   */


static void
generate_results(
		  PCL_CNT_TYPE * results_list, /* O: counter values */
		  result_descr_t results[PCL_MAX_COUNTER] /* I: result descr */
		  )

{
  int j;


  for (j = 0; j < PCL_MAX_COUNTER; ++j)
    {
      if (results[j].used)
	{
	  /* counter that contributes to result values */
	  if (results[j].operation == 1)
	    /* add to result value the value of the j-th low-level counter */
	    results_list[results[j].result_index] += results[j].result_value;
	  else
	    /* subtract from result value the value of the
	       j-th low-level counter
	    */
	    results_list[results[j].result_index] -= results[j].result_value;
	}
    }
}


/*                                                                            */
/*============================================================================*/
/* start couting.                                                             */
/* In this function, counting mode is checked.                                */


static int
start_counting(
		void
)

{
#if defined(COUNTER_SIZE_32)           /* 32-bit performance counters */
  struct itimerval itimer;


  /* set up signal handler */
  if ((old_signal_handler = signal(SIGVTALRM, interrupt_handling)) == SIG_ERR)
    return PCL_FAILURE;

  /* set interval timer to avoid overflows. */
  itimer.it_interval.tv_sec = 1;
  itimer.it_interval.tv_usec = 0;
  itimer.it_value.tv_sec = 1;
  itimer.it_value.tv_usec = 0;
  if (setitimer(ITIMER_VIRTUAL, &itimer, NULL) != 0)
    return PCL_FAILURE;
#endif


  /*--------------------------------------------------------------------------*/
  /* system specific actions */

#if defined(ALPHA_CRAY)

  {
    int sel0, sel1, sel2;	       /* counter selectors */
    unsigned int mode;


    /* initialize selectors with default count events (cycles,
       non_issue_cycles, long_stalls) */
    sel0 = sel1 = sel2 = 0;

    /* Extract selected events */
    if(results[0].used)
      sel0 = results[0].command;
    if(results[1].used)
      sel1 = results[1].command;
    if(results[2].used)
      sel2 = results[2].command;

    /* select count mode */
    if(count_mode == PCL_MODE_USER)
      mode = 0x01;
    else if(count_mode == PCL_MODE_SYSTEM)
      mode = 0x02;
    else
      mode = 0x03;

    /* Clear and start the counters */
    if(perfctr_start(&sel0, &sel1, &sel2, mode) != 0)
      return PCL_FAILURE;
    else
      return PCL_SUCCESS;
  }


  /*--------------------------------------------------------------------------*/

#elif defined(ALPHA_DEC)

  {
    long arg, rc, sel0, sel1, sel2;
    union pmctrs_ev5 pm_ev5;


    /* initialize selectors with some existing count events */
    sel0 = PF5_MUX0_CYCLES;
    sel1 = PF5_MUX1_NON_ISSUE_CYCLES;
    sel2 = PF5_MUX2_C_LONG_STALLS;

    /* Extract selected events */
    if(results[0].used)
      sel0 = results[0].command;
    if(results[1].used)
      sel1 = results[1].command;
    if(results[2].used)
      sel2 = results[2].command;


    /* Try to open a channel to the pfcntr driver (if not already opened) */
    if(fd == -1)
      {
	unsigned long cpu_mask;
	long cpu_num;

	/* get actual CPU we are running on */
	if(getsysinfo(GSI_CURRENT_CPU, &cpu_num, sizeof(cpu_num),
		      NULL, NULL, NULL) < 1)
	  return PCL_FAILURE;

	/* bind process to processor */
	cpu_mask = 1 << cpu_num;
	if (bind_to_cpu(getpid(), cpu_mask, BIND_NO_INHERIT) == -1)
	  return PCL_FAILURE;

	/* open device (only for this processor) */
	fd = open("/dev/pfcntr", O_RDONLY | PCNTOPENONE,
		  S_IRUSR | S_IRGRP | S_IROTH);
	if (fd == -1)
	  return PCL_FAILURE;
      }

    /* select items (= counters) */
    arg = PFM_COUNTERS;
    rc = ioctl(fd, PCNTSETITEMS, &arg);
    if (rc < 0)
      return PCL_FAILURE;

    /* set contexts in which to count */
    arg = PF5_CTXT_THIS_PROC;
    /* can anybody explain to me why the constants PF5_CTXT_... have a meaning
       different to that what their name suggests?
       If you want an explanation for the settings, check how it is done on
       the CRAY.
       */
    if(count_mode == PCL_MODE_USER)
      /* user mode only: 0x40000200 */
      arg |= PF5_CTXT_USER_AND_KERNEL;
    else if(count_mode == PCL_MODE_SYSTEM)
      /* kernel mode only: 0x40000100 */
      arg |= PF5_CTXT_PAL;
    else
      /* user or kernel mode: 0x00000000 */
      arg |= PF5_CTXT_ALL_MODES;
    rc = ioctl(fd, PCNT5CTXTS, &arg);
    if (rc < 0)
      return PCL_FAILURE;

    /* clear counters */
    rc = ioctl(fd, PCNTCLEARCNT);
    if (rc < 0)
      return PCL_FAILURE;

    /* select events */
    arg = sel0 | sel1 | sel2;
    rc = ioctl(fd, PCNT5MUX, &arg);
    if (rc < 0)
      return PCL_FAILURE;

    /* select interrupt frequency */
    arg = PF5_C0_INT_EVERY_65536
        | PF5_C1_INT_EVERY_65536
        | PF5_C2_INT_EVERY_16384;
    rc = ioctl(fd, PCNT5FREQ, &arg);
    if (rc < 0)
      return PCL_FAILURE;

    /* zero and restart selected counters */
    pm_ev5.pmctrs_ev5_cpu = PMCTRS_ALL_CPUS;
    pm_ev5.pmctrs_ev5_select = PF5_SEL_COUNTER_0
                             | PF5_SEL_COUNTER_1
                             | PF5_SEL_COUNTER_2;
    rc = ioctl(fd, PCNT5RESTART, &pm_ev5.pmctrs_ev5_long);
    if (rc < 0)
      return PCL_FAILURE;


    return PCL_SUCCESS;
  }


  /*--------------------------------------------------------------------------*/

#elif defined(MIPS)

  {
    int sel0, sel1;


    /* check correct counting mode */
    if(count_mode != PCL_MODE_USER)
      return PCL_MODE_NOT_SUPPORTED;

    /* initialize selectors with default count events (cycles, cycles) */
    sel0 = sel1 = 0;

    /* Extract selected events */
    if(results[0].used)
      sel0 = results[0].command;
    if(results[1].used)
      sel1 = results[1].command;

    /* Clear and start the counters */
    generation_number = start_counters(sel0, sel1);
    if (generation_number < 0)
      return PCL_FAILURE;

    return PCL_SUCCESS;
  }


  /*--------------------------------------------------------------------------*/

#elif defined(SPARC)

  {
    int rc, sel0, sel1;
    unsigned long long set_count, val;
    processorid_t proc;


    /* initialize selectors with default count events (cycles, cycles) */
    sel0 = sel1 = 0;

    /* Extract selected events */
    if(results[0].used)
      sel0 = results[0].command;
    if(results[1].used)
      sel1 = results[1].command;

    /* if counter device is not already opened, open device */
    if (fd == -1)
      {
	/* Try to open a channel to the perfmon driver */
	fd = open("/dev/perfmon", O_RDONLY);
	if (fd == -1)
	  return PCL_FAILURE;

	/* query on what processor we are running */
	if (processor_bind(P_LWPID, P_MYID, PBIND_QUERY, &proc) == -1)
	  return PCL_FAILURE;

	/* if not assigned take CPU 0 */
	if (proc == -1)
	  proc = 0;

	/* bind process to processor */
	if (processor_bind(P_LWPID, P_MYID, proc, NULL) == -1)
	  return PCL_FAILURE;
      }

    /* Set PCR to accumulate desired events */
    set_count = sel0 | sel1;
    if(count_mode == PCL_MODE_USER)
      /* user mode only */
      set_count |= PCR_USER_TRACE;
    else if(count_mode == PCL_MODE_SYSTEM)
      /* kernel mode only */
      set_count |= PCR_SYS_TRACE;
    else
      /* user or kernel mode */
      set_count |= (PCR_SYS_TRACE | PCR_USER_TRACE);
    rc = ioctl(fd, PERFMON_SETPCR, &set_count);
    if (rc < 0)
      return PCL_FAILURE;

    /* Clear and start counter */
    clr_pic();
    cpu_sync();

    /* read out counter values (used as base value) */
    val = get_pic();
    pic0_value = val & 0xffffffff;
    pic1_value = val >> 32;

    return PCL_SUCCESS;
  }


  /*--------------------------------------------------------------------------*/

#elif defined(POWERPC)

  {
    int i;
    struct PM_ENCODING_CONTROL uic;
    struct PM_MACHINE_CONTROL mc;
    union PM_RETN retn;


    /* Initialize control blocks to the default value, 0, for each field.  */
    /* This routine clears all fields in uic,sc,pi,pc to present the defaults.*/
    memset(&uic, 0, sizeof(uic));
    memset(&sc, 0, sizeof(sc));
    memset(&pi, 0, sizeof(pi));
    memset(&pc, 0, sizeof(pc));


    /* request encoding */
    /* Revision */
    uic.revision = PM_ENCODING_CONTROL_REV;
    /* list of requested events */
    uic.uel = &uel;
    /* corresponding list for requested events */
    uic.ueivl = &uivl;
    /* No event chain history control list */
    uic.uechl = NULL;
    /* No event counter control list */
    uic.uecntl = NULL;
    /* No event group control list */
    uic.uegrpl = NULL;

    /* Count in problem state */
    if(count_mode == PCL_MODE_USER)
      uic.c_fields.b.MACH_STATE = MACH_STATE_PROB;
    else if(count_mode == PCL_MODE_SYSTEM)
      uic.c_fields.b.MACH_STATE = MACH_STATE_SUPERV;
    else
      uic.c_fields.b.MACH_STATE = MACH_STATE_ANY;

    /* Count regardless of the state of the MM bit in the MSR */
    uic.c_fields.b.PMBIT = PMBIT_ANY;

    /* event list encoding */
    /* Count PCL_MAX_COUNTER events */
    uel.n = PCL_MAX_COUNTER;
    for(i = 0; i < PCL_MAX_COUNTER; ++i)
      {
	/* not PM_NOTHING! otherwise we get the wrong order of results */
	uel.e[i] = (results[i].used) ? results[i].command : default_event[i];
      }

    /* start encoding */
    retn.w = PMencode(&uic, &mc);


    /* process init */
    /* Set revision field */
    pi.revision = PM_PROCESS_INIT_REV;
    /* Post caller on each interrupt */
    pi.pc_flags.b.POSTMODE = PM_POSTMODE_INTERRUPT;
    /* Automatically count next set of events on each interrupt */
    pi.pc_flags.b.AUTOMODE = PM_AUTOMODE_NEXT;
    /* Don't start counting on PMprocessInit call */
    pi.pc_flags.b.INITNOSTART = PM_INIT_NOSTART;
    /* Continue counting forever */
    pi.tp = 1;


    /* Initialize for processing */
    retn.w = PMprocessInit(&pi, &mc);
    if (retn.w != 0)
      return PCL_FAILURE;


    /* set processor control */
    /* Set revision field */
    pc.revision = PM_PROCESS_CONTROL_REV;
    /* Start counting command */
    pc.command = PM_COUNT_START;


    /* status control */
    /* set revision field */
    sc.revision = PM_STATUS_CONTROL_REV;
    /* control command */
    sc.command = PM_STATUS_OPTION_ADJUST;
    /* No list of requested events */
    sc.elp = NULL;
    sc.eip = &eil;
    sc.sip = &ir;
    sc.tip = NULL;


    /* Start counting */
    retn.w = PMprocess(&pc);
    if (retn.w != 0)
      return PCL_FAILURE;

    return PCL_SUCCESS;
  }


  /*--------------------------------------------------------------------------*/

#elif defined(PENTIUM_LINUX)

  {
    int sel0, sel1, cpl;


    if (!initialized)
      {
	/* do it only once */
	initialized = 1;

	/* determine CPU type we are running on */
	if (determine_pentium_type() != 0)
	  return PCL_NOT_SUPPORTED;
      }

    /* initialize selectors with default count events (something existing) */
    if(event_table == event_table_pentium)
      sel0 = sel1 = P5_MEM_DATA_READ;
    else
      sel0 = sel1 = P6_DCU_LINES_IN;

    /* Extract selected events */
    if(results[0].used)
      sel0 = results[0].command;
    if(results[1].used)
      sel1 = results[1].command;

    /* check count mode */
    if(count_mode == PCL_MODE_USER)
      /* count only in user mode */
      cpl = 2;
    else if(count_mode == PCL_MODE_SYSTEM)
      /* count only system mode */
      cpl = 1;
    else
      /* count in system or user mode */
      cpl = 3;


    /* get actual clock value (to measure cycles) */
    clock_value = rdtsc();

    /* Clear and start the counters */
    start_counters(sel0, cpl, sel1, cpl);


    return PCL_SUCCESS;
  }


  /*--------------------------------------------------------------------------*/

#elif defined(PENTIUM_NT)

  /* ??? not implemented yet */


  /*--------------------------------------------------------------------------*/

#else
#error "unsupported system"
#endif
}


/*                                                                            */
/*============================================================================*/
/* stop couting.                                                              */


static int
stop_counting(
	       PCL_CNT_TYPE * results_list,	/* O: counter values */
	       int ncounter	       /* I: number of event counters */
)

{
  int i;


  /* check number of counters */
  if((ncounter < 1) || (ncounter >= PCL_MAX_EVENT_PER_CALL))
    return PCL_TOO_MANY_EVENTS;

  /* initialize results_list */
  for (i = 0; i < ncounter; ++i)
    results_list[i] = 0;


  /*--------------------------------------------------------------------------*/
  /* system specific actions */

#if defined(ALPHA_CRAY)

  {

    long ctr0, ctr1, ctr2;


    /* initialize values */
    ctr0 = ctr1 = ctr2 = 0;

    /* stop and read out performance counters */
    if(perfctr_stopread(&ctr0, &ctr1, &ctr2) != 0)
      return PCL_FAILURE;

    /* store result values */
    results[0].result_value = ctr0;
    results[1].result_value = ctr1;
    results[2].result_value = ctr2;

    /* generate results */
    generate_results(results_list, results);

    return PCL_SUCCESS;
  }


  /*--------------------------------------------------------------------------*/

#elif defined(ALPHA_DEC)

  {
    int rc;
    struct pfcntrs_ev5 cntrs_ev5[NCPUS];
    struct pfcntrs_ev5 *pfcntrs_ev5 = &cntrs_ev5[0];
    union pmctrs_ev5 pmctrs_ev5;


    /* get values of counters */
    pmctrs_ev5.pmctrs_ev5_long = 0;
    rc = ioctl(fd, PCNT5GETCNTRS, &pmctrs_ev5.pmctrs_ev5_long);
    if (rc < 0)
      return PCL_FAILURE;

    /* get interrupt values */
    rc = ioctl(fd, PCNT5GETCNT, &pfcntrs_ev5);
    if (rc < 0)
      return PCL_FAILURE;

    /* extract counter values (timer_frequency * interrupts + counter_values) */
    results[0].result_value = (cntrs_ev5[0].pf_cntr0 << 16)
                              + pmctrs_ev5.pmctrs_ev5_cntr0;
    results[1].result_value = (cntrs_ev5[0].pf_cntr1 << 16)
                              + pmctrs_ev5.pmctrs_ev5_cntr1;
    results[2].result_value = (cntrs_ev5[0].pf_cntr2 << 14)
                              + pmctrs_ev5.pmctrs_ev5_cntr2;

    /* generate results */
    generate_results(results_list, results);

    return PCL_SUCCESS;
  }


  /*--------------------------------------------------------------------------*/

#elif defined(MIPS)

  {
    PCL_CNT_TYPE c0, c1;	       /* counts */
    int res, e0, e1;		       /* counter selectors */


    /* Extract selected events */
    if(results[0].used == 1)
       e0 = results[0].command;
    else
       e0 = 0;
    if(results[1].used == 1)
       e1 = results[1].command;
    else
       e1 = 0;

    /* read hardware counters */
    res = read_counters(e0, &c0, e1, &c1);
    if ((res < 0) || (res != generation_number))
      return PCL_FAILURE;

    /* store result values */
    results[0].result_value = c0;
    results[1].result_value =  c1;

    /* generate results */
    generate_results(results_list, results);

    /* no error occured */
    return PCL_SUCCESS;
  }

  /*--------------------------------------------------------------------------*/

#elif defined(SPARC)

  {
    unsigned long long val;

    /* if (for some reason) the performance counter device is not open,
       signal an error
    */
    if (fd == -1)
      return PCL_FAILURE;

    /* finish all outstanding operations */
    cpu_sync();

    /* get PIC-values as a difference between the actual value and the value
       gotten when starting the counts. The lower 32 bits are counter0 results,
       the upper 32 bits are counter1 results.
    */
    val = get_pic();
    results[0].result_value = (val & 0xffffffff) - pic0_value;
    results[1].result_value =  (val >> 32) - pic1_value;

    /* generate results */
    generate_results(results_list, results);

    /* no error occured */
    return PCL_SUCCESS;
  }


  /*--------------------------------------------------------------------------*/

#elif defined(POWERPC)

  {
    union PM_RETN retn;


    /* Read the accumulated values and store for further operations. */
    retn.w = PMstatus(&sc);
    if (retn.w != 0)
      return PCL_FAILURE;

    /* store result values */
    results[0].result_value = eil.ei[0].curr;
    results[1].result_value = eil.ei[1].curr;
    results[2].result_value = eil.ei[2].curr;
    results[3].result_value = eil.ei[3].curr;

    /* Terminate counting */
    /* Set revision field */
    pc.revision = PM_PROCESS_CONTROL_REV;
    pc.command = PM_COUNT_TERM;
    PMprocess(&pc);

    /* generate results */
    generate_results(results_list, results);

    return PCL_SUCCESS;
  }


  /*--------------------------------------------------------------------------*/

#elif defined(PENTIUM_LINUX)

  {
    long long c0, c1, clocks;
    double timer;


    /* read counter values */
    read_counters(&c0, &c1, &timer, &clocks);

    /* get actual cycle count */
    clocks = rdtsc() - clock_value;

    /* store result values */
    results[0].result_value = c0;
    results[1].result_value = c1;

    /* special case: cycle count */
    if(results[0].event == PCL_CYCLES)
      results[0].result_value = clocks;
    else if(results[1].event == PCL_CYCLES)
      results[1].result_value = clocks;

    /* generate results */
    generate_results(results_list, results);

    /* no error occured */
    return PCL_SUCCESS;
  }


  /*--------------------------------------------------------------------------*/

#elif defined(PENTIUM_NT)
  /* ??? */


  /*--------------------------------------------------------------------------*/

#error "unsupported system"


  /*--------------------------------------------------------------------------*/

#else
#error "unsupported system"
#endif
}


/*                                                                            */
/*============================================================================*/
/* check input arguments to PCL_query and PCL_start                           */


static int
check_and_init(
	       int *counter_list,     /* I: event counters to be started */
	       int ncounter,	       /* I: number of event counters */
	       unsigned int flags     /* I: mode flags (PCL_MODE_xxx) */
	       )

{
  int i;


  /*--------------------------------------------------------------------------*/
  /* initialization (needed on Pentiums at this place to get the right table) */

#if defined(PENTIUM_LINUX)
  if (!initialized)
    {
      /* do it only once */
      initialized = 1;

      /* determine CPU type (needed for correct event table) */
      if (determine_pentium_type() != 0)
	return PCL_NOT_SUPPORTED;
    }
#endif


  /*--------------------------------------------------------------------------*/
  /* check call arguments */

  /* check that number of requested events is smaller than maximum allowed
     number */
  if (ncounter >= PCL_MAX_EVENT_PER_CALL)
    return PCL_TOO_MANY_EVENTS;

  /* check that event identifiers are in allowed range */
  for (i = 0; i < ncounter; ++i)
    if ((counter_list[i] < 0) || (counter_list[i] >= PCL_MAX_EVENT))
      return PCL_ILL_EVENT;

  /* check correct mode */
  if((flags != PCL_MODE_USER)
     && (flags != PCL_MODE_SYSTEM)
     && (flags != PCL_MODE_USER_SYSTEM))
    return PCL_MODE_NOT_SUPPORTED;

  /* check if count mode is supported */
  if((flags & allowed_count_mode) != flags)
    return PCL_MODE_NOT_SUPPORTED;


  /* everything ok */
  return PCL_SUCCESS;
}


/*                                                                            */
/*============================================================================*/
/* query for functionality.                                                   */
/* returns PCL_SUCCESS if requested event counting is possible                */


/* Fortran version */
int
fname(pclquery, PCLQUERY)(
	   int *counter_list,	       /* I: requested event counters */
	   int *ncounter,	       /* I: # of requested event counters */
	   unsigned int *flags          /* I: mode flags (PCL_MODE_xxx) */
)

{
  return PCLquery(counter_list, *ncounter, *flags);
}


int
PCLquery(
	   int *counter_list,	       /* I: requested event counters */
	   int ncounter,	       /* I: # of requested event counters */
	   unsigned int flags          /* I: mode flags (PCL_MODE_xxx) */
)

{
  int res;
  result_descr_t tmp_results[PCL_MAX_COUNTER];


  /* check call arguments */
  if((res = check_and_init(counter_list, ncounter, flags)) != PCL_SUCCESS)
    return res;

  /* try to get a mapping of events to counters (with tmp_results) */
  return get_low_level_counters(counter_list, ncounter, tmp_results);
}


/*                                                                            */
/*============================================================================*/
/* start performance counting                                                 */
/* returns PCL_SUCCESS if everything went ok                                  */
/* If this is a nested call we do the following:                              */
/*    - increment the nesting level                                           */
/*    - check that same events are requested than in previous level           */
/*    - stops previous couting and store result values in <partial_results>   */


/* Fortran version */
int
fname(pclstart, PCLSTART)(
	   int *counter_list,	       /* I: event counters to be started */
	   int *ncounter,	       /* I: number of event counters */
	   unsigned int *flags          /* I: mode flags (PCL_MODE_xxx) */
)

{
  return PCLstart(counter_list, *ncounter, *flags);
}


int
PCLstart(
	   int *counter_list,	       /* I: event counters to be started */
	   int ncounter,	       /* I: number of event counters */
	   unsigned int flags          /* I: mode flags (PCL_MODE_xxx) */
)

{
  int i, res;
  PCL_CNT_TYPE results_list[PCL_MAX_EVENT_PER_CALL];


  /*--------------------------------------------------------------------------*/
  /* check call arguments */

  if((res = check_and_init(counter_list, ncounter, flags)) != PCL_SUCCESS)
    return res;

  /* increment call nesting */
  if (nesting_level + 1 >= PCL_MAX_NESTING_LEVEL)
    return PCL_TOO_MANY_NESTINGS;


  /*--------------------------------------------------------------------------*/
  /* handle previous calls */

  /* check that the same events are requested as in lower nesting levels and
     stop couting for previous call */
  if (nesting_level >= 0)
    {
      /* check number of requests */
      if (ncounter != event_count)
	return PCL_ILL_NESTING;

      /* check that event requests of this call are the same than that of
         previous call */
      for (i = 0; i < ncounter; ++i)
	if (counter_list[i] != events[i])
	  return PCL_ILL_NESTING;

      /* check correct mode */
      if(flags != count_mode)
	return PCL_MODE_NOT_SUPPORTED;

      /* stop counting for previous call; counting is continued when PCL_stop
         of this call is finished */
      res = stop_counting(results_list, ncounter);
      if(res != PCL_SUCCESS)
	return res;

      /* add results to partial results of previous call */
      for (i = 0; i < ncounter; ++i)
	partial_results[nesting_level][i] += results_list[i];
    }

  else
    {
      /* store count mode */
      count_mode = flags;

      /* store request for this call */
      event_count = ncounter;
      for (i = 0; i < ncounter; ++i)
	events[i] = counter_list[i];
    }


  /* increment nesting level */
  ++nesting_level;

  /* zero partial results for this nesting level */
  for (i = 0; i < PCL_MAX_EVENT_PER_CALL; ++i)
    partial_results[nesting_level][i] = 0;

  /* try to get a mapping of events to counters */
  if((res = get_low_level_counters(counter_list, ncounter, results)) !=
     PCL_SUCCESS)
    return res;

  /* start the counters */
  return start_counting();
}


/*                                                                            */
/*============================================================================*/
/* reads performance counters                                                 */
/* result_list must be a pointer to a sufficiently large buffer.              */
/* returns PCL_SUCCESS if everything went ok                                  */


/* Fortran version */
int
fname(pclread, PCLREAD)(
	  PCL_CNT_TYPE * results_list, /* O: counter values */
	  int *ncounter		       /* I: number of event counters */
)

{
  return PCLread(results_list, *ncounter);
}


int
PCLread(
	  PCL_CNT_TYPE * results_list, /* O: counter values */
	  int ncounter		       /* I: number of event counters */
)

{
  int i, res;


  /*--------------------------------------------------------------------------*/
  /* check call arguments */

  /* check call level */
  if (nesting_level < 0)
    return PCL_ILL_NESTING;

  /* check requests for this call with corresponding PCL_start */
  if (event_count != ncounter)
    return PCL_ILL_NESTING;


  /*--------------------------------------------------------------------------*/
  /* stop and start counters */

  /* stop counting and read values */
  if ((res = stop_counting(results_list, ncounter)) != PCL_SUCCESS)
    return res;

  /* add to partial results */
  for (i = 0; i < ncounter; ++i)
    partial_results[nesting_level][i] += results_list[i];

  /* restart counting */
  return start_counting();
}


/*                                                                            */
/*============================================================================*/
/* stops performance counting and reads counters values.                      */
/* result_list must be a pointer to a sufficiently large buffer.              */
/* returns PCL_SUCCESS if everything went ok                                  */


/* Fortran version */
int
fname(pclstop, PCLSTOP)(
	  PCL_CNT_TYPE * results_list, /* O: counter values */
	  int *ncounter		       /* I: number of event counters */
)

{
  return PCLstop(results_list, *ncounter);
}


int
PCLstop(
	  PCL_CNT_TYPE * results_list, /* O: counter values */
	  int ncounter		       /* I: number of event counters */
)

{
  int i, res;
  PCL_CNT_TYPE tmp_results_list[PCL_MAX_EVENT_PER_CALL];
#if defined(COUNTER_SIZE_32)           /* 32-bit performance counters */
  struct itimerval itimer;
#endif


  /*--------------------------------------------------------------------------*/
  /* check call arguments */

  /* check call level */
  if (nesting_level < 0)
    return PCL_ILL_NESTING;

  /* check requests for this call with corresponding PCL_start */
  if (event_count != ncounter)
    return PCL_ILL_NESTING;


  /*--------------------------------------------------------------------------*/
  /* stop counting and read values */

  res = stop_counting(tmp_results_list, ncounter);
  if(res != PCL_SUCCESS)
    return res;

  /* add with partial results gotten with PCL_read */
  for (i = 0; i < ncounter; ++i)
    {
      partial_results[nesting_level][i] += tmp_results_list[i];
      results_list[i] = partial_results[nesting_level][i];
    }


  /*--------------------------------------------------------------------------*/
  /* update data for previous calls */

  /* decrement nesting level */
  --nesting_level;

  if (nesting_level >= 0)
    {
      /* add results to partial results of previous call */
      for (i = 0; i < ncounter; ++i)
	partial_results[nesting_level][i] += results_list[i];

      /* restart call of previous level which has been stopped when PCL_start
         of this level was executed */
      res = start_counting();
      if(res != PCL_SUCCESS)
	return res;
    }


  /*--------------------------------------------------------------------------*/
  /* stop time interrupts (overflow avoidance) */

#if defined(COUNTER_SIZE_32)           /* 32-bit performance counters */
  if (nesting_level == -1)
    {

      /* set up signal handler */
      if (signal(SIGVTALRM, old_signal_handler) == SIG_ERR)
	return PCL_FAILURE;

      /* cancel interval timer */
      itimer.it_interval.tv_sec = 0;
      itimer.it_interval.tv_usec = 0;
      itimer.it_value.tv_sec = 0;
      itimer.it_value.tv_usec = 0;
      setitimer(ITIMER_VIRTUAL, &itimer, NULL);
    }
#endif


  /*--------------------------------------------------------------------------*/
  /* leave funtion */

  return PCL_SUCCESS;
}


/*============================================================================*
 *                             that's all folks                               *
 *============================================================================*/
