/*==============================================================================


   Purpose:    interface for Performance Counter Library (PCL)


   Author:     Heinz Ziegler, Rudolf Berrendorf
               Research Centre Juelich
               Central Institute for Applied Mathematics
               D-52425 Juelich
               Tel: ++49 2461 61-3218
               r.berrendorf@fz-juelich.de


   last mod:   14.10.1998


   SUPPORT AND DISCLAIMER

   The software in this file (the "Software") is an unsupported product and is
   provided by Zentralinstitut fuer Angewandte Mathematik, Forschungszentrum
   Juelich GmbH (ZAM) "AS IS". We welcome comments, suggestions, and problem
   reports and ask that you mail those to zam@fz-juelich.de; however, we
   make no promises to respond to any item.

   ZAM makes no warrranties or representations, expressed or implied, including,
   but not limited to, implied warranties of merchantability and fitness for a
   particular purpose, as to any element of the Software or any support
   provided in connection with the Software. In no event shall ZAM be
   responsible for any damages, including but not limited to consequential
   damages, arising from or relating to any use of the Software or related
   support.


   COPYRIGHT

   (c) COPYRIGHT 1998
   Zentralinstitut fuer Angewandte Mathematik, Forschungszentrum Juelich GmbH

   The Software is copyrighted. For research purposes, you can use it for free.
   If you plan to use it commercially, please contact zam@fz-juelich.de.


==============================================================================*/

#if !defined(PCL_INCLUDED)
#define PCL_INCLUDED


/*                                                                            */
/*============================================================================*/
/* define exactly one of the following systems:                               */


/* you can do that either here or in the Makefile with -Dsystem_name */
/*#define ALPHA_CRAY */
/*#define ALPHA_DEC */
/*#define MIPS */
/*#define SPARC */
/*#define POWERPC */
/*#define PENTIUM_LINUX */
/*#define PENTIUM_NT */


/*                                                                            */
/*============================================================================*/
/* Include hardware/os dependend libraries                                    */


#include <stdio.h>
#include <stdlib.h>


/*                                                                            */
/*============================================================================*/
/* Generate Fortran-names out of C-names                                      */


#if defined(ALPHA_DEC) || defined(MIPS) || defined(SPARC) || defined(PENTIUM_LINUX)
/* all lowercase, underscore appended */
#define fname(n1, n2) n1##_
#elif defined(ALPHA_CRAY)
/* uppercase letters */
#define fname(n1, n2) n2
#else
/* all lowercase */
#define fname(n1, n2) n1
#endif


/*                                                                            */
/*============================================================================*/
/* some constants:                                                            */


/* maximum values */
#define PCL_MAX_EVENT          61      /* number of supported events */
#define PCL_MAX_EVENT_PER_CALL  4      /* max. number of events per call */
#define PCL_MAX_NESTING_LEVEL  16      /* max. number of nested calls to PCL */

/* return codes */
#define PCL_SUCCESS             0      /* successful operation */
#define PCL_NOT_SUPPORTED      -1      /* failure: event not supported */
#define PCL_TOO_MANY_EVENTS    -2      /* failure: not all requests supported */
#define PCL_TOO_MANY_NESTINGS  -3      /* too many call nesting */
#define PCL_ILL_NESTING        -4      /* illegal call nesting */
#define PCL_ILL_EVENT          -5      /* illegal event identifier */
#define PCL_MODE_NOT_SUPPORTED -6      /* mode not supported */
#define PCL_FAILURE            -7      /* failure: something else */


/* result type for performance counters (64 bit integer) */
#if defined(ALPHA_CRAY)
#define PCL_CNT_TYPE long

#elif defined(ALPHA_DEC)
#define PCL_CNT_TYPE long

#elif defined(MIPS)
#define PCL_CNT_TYPE long long

#elif defined(SPARC)
#define PCL_CNT_TYPE long long

#elif defined(POWERPC)
#define PCL_CNT_TYPE long long

#elif defined(PENTIUM_LINUX)
#define PCL_CNT_TYPE long long

#elif defined(PENTIUM_NT)
#define PCL_CNT_TYPE long long

#else
#error "unsupported system"

#endif


/*                                                                            */
/*============================================================================*/
/* mode flags:                                                                */


/* define what to count */
#define PCL_MODE_USER         1   /* count for user mode only */
#define PCL_MODE_SYSTEM       2   /* count for system mode only */
#define PCL_MODE_USER_SYSTEM  (PCL_MODE_USER|PCL_MODE_SYSTEM)  /* count both */


/*                                                                            */
/*============================================================================*/
/* supported events:                                                          */


/* Level-1-Cache */
#define PCL_L1CACHE_READ            0
#define PCL_L1CACHE_WRITE           1
#define PCL_L1CACHE_READWRITE       2
#define PCL_L1CACHE_HIT             3
#define PCL_L1CACHE_MISS            4
#define PCL_L1CACHE_USED_LINES      5

/* Level-1-Data-Cache */
#define PCL_L1DCACHE_READ           6
#define PCL_L1DCACHE_WRITE          7
#define PCL_L1DCACHE_READWRITE      8
#define PCL_L1DCACHE_HIT            9
#define PCL_L1DCACHE_MISS           10
#define PCL_L1DCACHE_USED_LINES     11

/* Level-1-Instruction-Cache */
#define PCL_L1ICACHE_READ           12
#define PCL_L1ICACHE_WRITE          13
#define PCL_L1ICACHE_READWRITE      14
#define PCL_L1ICACHE_HIT            15
#define PCL_L1ICACHE_MISS           16
#define PCL_L1ICACHE_USED_LINES     17

/*----------------------------------------------------------------------------*/

/* Level-2-Cache */
#define PCL_L2CACHE_READ            18
#define PCL_L2CACHE_WRITE           19
#define PCL_L2CACHE_READWRITE       20
#define PCL_L2CACHE_HIT             21
#define PCL_L2CACHE_MISS            22
#define PCL_L2CACHE_USED_LINES      23

/* Level-2-Data-Cache */
#define PCL_L2DCACHE_READ           24
#define PCL_L2DCACHE_WRITE          25
#define PCL_L2DCACHE_READWRITE      26
#define PCL_L2DCACHE_HIT            27
#define PCL_L2DCACHE_MISS           28
#define PCL_L2DCACHE_USED_LINES     29

/* Level-2-Instruction-Cache */
#define PCL_L2ICACHE_READ           30
#define PCL_L2ICACHE_WRITE          31
#define PCL_L2ICACHE_READWRITE      32
#define PCL_L2ICACHE_HIT            33
#define PCL_L2ICACHE_MISS           34
#define PCL_L2ICACHE_USED_LINES     35

/*----------------------------------------------------------------------------*/

/* TLB */
#define PCL_TLB_HIT                 36
#define PCL_TLB_MISS                37

/* Instruction-TLB */
#define PCL_ITLB_HIT                38
#define PCL_ITLB_MISS               39

/* Data-TLB */
#define PCL_DTLB_HIT                40
#define PCL_DTLB_MISS               41

/*----------------------------------------------------------------------------*/

/* Cycles */
#define PCL_CYCLES                  42

/* Operations */
#define PCL_INTEGER_INSTR           43
#define PCL_FP_INSTR                44
#define PCL_LOAD_INSTR              45
#define PCL_STORE_INSTR             46
#define PCL_LOADSTORE_INSTR         47
#define PCL_INSTR                   48

/* Jump Instructions */
#define PCL_JUMP_SUCCESS            49
#define PCL_JUMP_UNSUCCESS          50
#define PCL_JUMP                    51

/* Atomic Instructions */
#define PCL_ATOMIC_SUCCESS          52
#define PCL_ATOMIC_UNSUCCESS        53
#define PCL_ATOMIC                  54

/*----------------------------------------------------------------------------*/

/* Stalls */
#define PCL_STALL_INTEGER           55
#define PCL_STALL_FP                56
#define PCL_STALL_JUMP              57
#define PCL_STALL_LOAD              58
#define PCL_STALL_STORE             59
#define PCL_STALL                   60


/*                                                                            */
/*============================================================================*/
/* C++:                                                                      */


#ifdef	__cplusplus
extern "C" {
#endif


/*                                                                            */
/*============================================================================*/
/* global variables:                                                          */


/* PCL version number */
extern char *PCL_version_number;

/*                                                                            */
/*============================================================================*/
/* interface functions:                                                       */


/* query for functionality.
   returns PCL_SUCCESS if requested event counting is possible
 */
int PCLquery(
	       int *counter_list,      /* I: requested event counters */
	       int ncounter,	       /* I: number of requested event
				          counters */
	       unsigned int flags      /* I: mode flags (PCL_MODE_xxx) */
);

/* Fortran version */
int
fname(pclquery, PCLQUERY)(
	   int *counter_list,	       /* I: requested event counters */
	   int *ncounter,	       /* I: # of requested event counters */
	   unsigned int *flags          /* I: mode flags (PCL_MODE_xxx) */
);


/* start performance counting
   returns PCL_SUCCESS if everything went ok
 */
int PCLstart(
	       int *counter_list,      /* I: event counters to be started */
	       int ncounter,	       /* I: number of event counters */
	       unsigned int flags      /* I: flags (PCL_MODE_xxx) */
);

/* Fortran version */
int
fname(pclstart, PCLSTART)(
	   int *counter_list,	       /* I: event counters to be started */
	   int *ncounter,	       /* I: number of event counters */
	   unsigned int *flags          /* I: mode flags (PCL_MODE_xxx) */
);


/* reads performance counters.
   result_list must be a pointer to a sufficiently large buffer.
   returns PCL_SUCCESS if everything went ok
 */
int PCLread(
	      PCL_CNT_TYPE * results_list,	/* O: counter values */
	      int ncounter	       /* I: number of event counters */
);

/* Fortran version */
int
fname(pclread, PCLREAD)(
	  PCL_CNT_TYPE * results_list, /* O: counter values */
	  int *ncounter		       /* I: number of event counters */
);


/* stops performance counting and reads counters values.
   result_list must be a pointer to a sufficiently large buffer.
   returns PCL_SUCCESS if everything went ok
 */
int PCLstop(
	      PCL_CNT_TYPE * results_list,	/* O: counter values */
	      int ncounter	       /* I: number of event counters */
);

/* Fortran version */
int
fname(pclstop, PCLSTOP)(
	  PCL_CNT_TYPE * results_list, /* O: counter values */
	  int *ncounter		       /* I: number of event counters */
);


/*                                                                            */
/*============================================================================*/
/*                                                                            */

#ifdef	__cplusplus
}
#endif


/*                                                                            */
/*============================================================================*/
/* C++:                                                                      */


#endif				       /* !defined(PCL_INCLUDED) */


/*============================================================================*
 *                             that's all folks                               *
 *============================================================================*/
