/*==============================================================================


   Purpose:    test program for Performance Counter Library (PCL)


   Author:     Heinz Ziegler, Rudolf Berrendorf
               Research Centre Juelich
               Central Institute for Applied Mathematics
               D-52425 Juelich
               Tel: ++49 2461 61-3218
               r.berrendorf@fz-juelich.de


   last mod:   13.10.1998


   SUPPORT AND DISCLAIMER

   The software in this file (the "Software") is an unsupported product and is
   provided by Zentralinstitut fuer Angewandte Mathematik, Forschungszentrum
   Juelich GmbH (ZAM) "AS IS". We welcome comments, suggestions, and problem
   reports and ask that you mail those to zam@fz-juelich.de; however, we
   make no promises to respond to any item.

   ZAM makes no warrranties or representations, expressed or implied, including,
   but not limited to, implied warranties of merchantability and fitness for a
   particular purpose, as to any element of the Software or any support
   provided in connection with the Software. In no event shall ZAM be
   responsible for any damages, including but not limited to consequential
   damages, arising from or relating to any use of the Software or related
   support.


   COPYRIGHT

   (c) COPYRIGHT 1998
   Zentralinstitut fuer Angewandte Mathematik, Forschungszentrum Juelich GmbH

   The Software is copyrighted. For research purposes, you can use it for free.
   If you plan to use it commercially, please contact zam@fz-juelich.de.


==============================================================================*/


#include "pcl.h"

#include <string.h>


/* dimension of test arrays */
#define N 200

/* output format and type */
#define OFORMAT   "%13.0f"
#define OTYPE     double

/* names of events */
static const char *(eventname[PCL_MAX_EVENT]) =
{
  "L1CACHE_READ",
	  "L1CACHE_WRITE",
	  "L1CACHE_READWRITE",
	  "L1CACHE_HIT",
	  "L1CACHE_MISS",
	  "L1CACHE_USED_LINES",
	  "L1DCACHE_READ",
	  "L1DCACHE_WRITE",
	  "L1DCACHE_READWRITE",
	  "L1DCACHE_HIT",
	  "L1DCACHE_MISS",
	  "L1DCACHE_USED_LINES",
	  "L1ICACHE_READ",
	  "L1ICACHE_WRITE",
	  "L1ICACHE_READWRITE",
	  "L1ICACHE_HIT",
	  "L1ICACHE_MISS",
	  "L1ICACHE_USED_LINES",
	  "L2CACHE_READ",
	  "L2CACHE_WRITE",
	  "L2CACHE_READWRITE",
	  "L2CACHE_HIT",
	  "L2CACHE_MISS",
	  "L2CACHE_USED_LINES",
	  "L2DCACHE_READ",
	  "L2DCACHE_WRITE",
	  "L2DCACHE_READWRITE",
	  "L2DCACHE_HIT",
	  "L2DCACHE_MISS",
	  "L2DCACHE_USED_LINES",
	  "L2ICACHE_READ",
	  "L2ICACHE_WRITE",
	  "L2ICACHE_READWRITE",
	  "L2ICACHE_HIT",
	  "L2ICACHE_MISS",
	  "L2ICACHE_USED_LINES",
	  "TLB_HIT",
	  "TLB_MISS",
	  "ITLB_HIT",
	  "ITLB_MISS",
	  "DTLB_HIT",
	  "DTLB_MISS",
	  "CYCLES",
	  "INTEGER_INSTR",
	  "FP_INSTR",
	  "LOAD_INSTR",
	  "STORE_INSTR",
	  "LOADSTORE_INSTR",
	  "INSTR",
	  "JUMP_SUCCESS",
	  "JUMP_UNSUCCESS",
	  "JUMP",
	  "ATOMIC_SUCCESS",
	  "ATOMIC_UNSUCCESS",
	  "ATOMIC",
	  "STALL_INTEGER",
	  "STALL_FP",
	  "STALL_JUMP",
	  "STALL_LOAD",
	  "STALL_STORE",
	  "STALL"
};

/* error messages */
static const char *(errornames[]) =
{
  "success",
	  "not supported",
	  "too many events",
	  "too many nestings",
	  "illegal nesting",
	  "illegal event",
	  "mode not supported",
	  "failure"
};


static void dummy(unsigned int ijk);

/*                                                                            */
/*============================================================================*/
/*                                                                            */


static void
matadd(double a[N][N], double b[N][N], double c[N][N])

{
  int i, j;


  for (i = 0; i < N; ++i)
    for (j = 0; j < N; ++j)
      a[i][j] = b[i][j] + c[i][j];
}


/*                                                                            */
/*============================================================================*/
/*                                                                            */


static void
usage(char **argv)

{
  printf("usage: %s [options]\n"
	 "where options might be:\n"
	 "\t-overhead      measure overhead for PCL\n"
	 "\t-startstop     do PCLstart/PCLstop\n"
	 "\t-read          do PCLstart/PCLread/PCLstop\n"
	 "\t-nested        do PCLstart/PCLstart/PCLstop/PCLstop\n"
	 "\t-multiple      do PCLstart(2 events)/PCLstop(2 events)\n"
	 "\t-overflow      generate 32-bit overflow\n"
	 "\t-mode (1|2|3)  user | system | user+system mode\n"
	 "\t-all           all of the above\n",
	 argv[0]);
  exit(1);
}


/*                                                                            */
/*============================================================================*/
/*                                                                            */


int
main(int argc, char **argv)
{
  int i, j, n = N;
  int do_overhead, do_startstop, do_read, do_nested, do_multiple, do_overflow;
  int res, event, counter_list[2];
  unsigned int flags;
  double a[N][N], b[N][N], c[N][N];
  PCL_CNT_TYPE result[2] = {0, 0};


  /* do counting in for user mode */
  flags = PCL_MODE_USER;

  /* parse arguments */
  if (argc > 1)
    {
      do_overhead =
	      do_startstop =
	      do_read =
	      do_nested =
	      do_multiple =
	      do_overflow = 0;

      for (i = 1; i < argc; ++i)
	{
	  /* parse options */
	  if (!strcmp(argv[i], "-overhead"))
	    do_overhead = 1;

	  else if (!strcmp(argv[i], "-startstop"))
	    do_startstop = 1;

	  else if (!strcmp(argv[i], "-read"))
	    do_read = 1;

	  else if (!strcmp(argv[i], "-nested"))
	    do_nested = 1;

	  else if (!strcmp(argv[i], "-multiple"))
	    do_multiple = 1;

	  else if (!strcmp(argv[i], "-overflow"))
	    do_overflow = 1;

	  else if (!strcmp(argv[i], "-mode"))
	    {
	      ++i;
	      if ((i >= argc) || (sscanf(argv[i], "%d", &flags) != 1))
		usage(argv);
	    }

	  else if (!strcmp(argv[i], "-all"))
	    do_overhead =
		    do_startstop =
		    do_read =
		    do_nested =
		    do_multiple =
		    do_overflow = 1;

	  else
	    usage(argv);
	}
    }

  else
    {
      do_overhead =
	      do_read =
	      do_nested =
	      do_multiple =
	      do_overflow = 0;
      do_startstop = 1;
    }


  /* initialize test arrays */
  for (i = 0; i < n; ++i)
    for (j = 0; j < n; ++j)
      a[i][j] = b[i][j] = c[i][j] = 1.0;


  /* do some work */
  matadd(a, b, c);


/*-----------------------------------------------------------*/
  /* overhead */

  if (do_overhead)
    {
      printf("overhead:\n");
      printf("---------\n");

      for (event = 0; event < PCL_MAX_EVENT; ++event)
	{
	  /* define events to be measured */
	  printf("%-20s: ", eventname[event]);
	  counter_list[0] = event;

	  /* query for functionality */
	  if ((res = PCLquery(counter_list, 1, flags)) == PCL_SUCCESS)
	    {
	      /* start counting */
	      if ((res = PCLstart(counter_list, 1, flags)) != PCL_SUCCESS)
		{
		  printf("problem with starting event %d (%s)\n",
			 event, errornames[-res]);
		  exit(1);
		}

	      /* start counting */
	      if ((res = PCLstart(counter_list, 1, flags)) != PCL_SUCCESS)
		{
		  printf("problem with starting event %d (%s)\n",
			 event, errornames[-res]);
		  exit(1);
		}

	      /* stop counting */
	      if ((res = PCLstop(result, 1)) != PCL_SUCCESS)
		printf("problem with stopping event %d (%s)\n",
		       event, errornames[-res]);

	      /* stop counting */
	      if ((res = PCLstop(result, 1)) != PCL_SUCCESS)
		printf("problem with stopping event %d (%s)\n",
		       event, errornames[-res]);
	      else
		printf(OFORMAT "\n", (OTYPE) result[0]);
	    }

	  else
	    printf("%s\n", errornames[-res]);
	}
    }


/*-----------------------------------------------------------*/
  /* test start/stop for all possible events */

  if (do_startstop)
    {
      printf("\ntesting start/stop for available events:\n");
      printf("----------------------------------------\n");

      for (event = 0; event < PCL_MAX_EVENT; ++event)
	{
	  /* define events to be measured */
	  printf("%-20s: ", eventname[event]);
	  counter_list[0] = event;

	  /* query for functionality */
	  if ((res = PCLquery(counter_list, 1, flags)) == PCL_SUCCESS)
	    {
	      /* start counting */
	      if ((res = PCLstart(counter_list, 1, flags)) != PCL_SUCCESS)
		{
		  printf("problem with starting event %d (%s)\n",
			 event, errornames[-res]);
		  exit(1);
		}

	      /* do work */
	      matadd(a, b, c);

	      /* stop counting */
	      if ((res = PCLstop(result, 1)) != PCL_SUCCESS)
		printf("problem with stopping event %d (%s)\n",
		       event, errornames[-res]);
	      else
		printf(OFORMAT "\n", (OTYPE) result[0]);
	    }

	  else
	    printf("%s\n", errornames[-res]);
	}
    }


/*-----------------------------------------------------------*/
  /* test read for all possible events */

  if (do_read)
    {
      printf("\ntesting read for available events:\n");
      printf("----------------------------------\n");

      for (event = 0; event < PCL_MAX_EVENT; ++event)
	{
	  /* define events to be measured */
	  printf("%-20s: ", eventname[event]);
	  counter_list[0] = event;

	  /* query for functionality */
	  if ((res = PCLquery(counter_list, 1, flags)) == PCL_SUCCESS)
	    {
	      /* start counting */
	      if ((res = PCLstart(counter_list, 1, flags)) != PCL_SUCCESS)
		{
		  printf("problem with starting event %d (%s)\n",
			 event, errornames[-res]);
		  exit(1);
		}

	      /* do work */
	      matadd(a, b, c);

	      /* read values */
	      if ((res = PCLread(&result[0], 1)) != PCL_SUCCESS)
		printf("problem with reading event %d (%s)\n",
		       event, errornames[-res]);

	      /* do work */
	      matadd(a, b, c);

	      /* stop counting */
	      if ((res = PCLstop(&result[1], 1)) != PCL_SUCCESS)
		printf("problem with stopping event %d (%s)\n",
		       event, errornames[-res]);
	      else
		printf(OFORMAT " " OFORMAT "\n",
		       (OTYPE) result[0], (OTYPE) result[1]);
	    }

	  else
	    printf("%s\n", errornames[-res]);
	}
    }


/*-----------------------------------------------------------*/
  /* test nested calls for all possible events */

  if (do_nested)
    {
      printf("\ntesting nested calls for available events:\n");
      printf("--------------------------------------------\n");

      for (event = 0; event < PCL_MAX_EVENT; ++event)
	{
	  /* define events to be measured */
	  printf("%-20s: ", eventname[event]);
	  counter_list[0] = event;

	  /* query for functionality */
	  if ((res = PCLquery(counter_list, 1, flags)) == PCL_SUCCESS)
	    {
	      /* start counting */
	      if ((res = PCLstart(counter_list, 1, flags)) != PCL_SUCCESS)
		{
		  printf("problem with starting event %d (%s)\n",
			 event, errornames[-res]);
		  exit(1);
		}

	      /* do work */
	      matadd(a, b, c);

	      /* start counting */
	      if ((res = PCLstart(counter_list, 1, flags)) != PCL_SUCCESS)
		{
		  printf("problem with starting event %d (%s)\n",
			 event, errornames[-res]);
		  exit(1);
		}

	      /* do work */
	      matadd(a, b, c);

	      /* stop counting */
	      if ((res = PCLstop(&result[0], 1)) != PCL_SUCCESS)
		printf("problem with stopping event %d (%s)\n",
		       event, errornames[-res]);

	      /* stop counting */
	      if ((res = PCLstop(&result[1], 1)) != PCL_SUCCESS)
		printf("problem with stopping event %d (%s)\n",
		       event, errornames[-res]);
	      else
		printf(OFORMAT " " OFORMAT "\n",
		       (OTYPE) result[0], (OTYPE) result[1]);
	    }

	  else
	    printf("%s\n", errornames[-res]);
	}
    }


/*-----------------------------------------------------------*/
  /* test for multiple events */

  if (do_multiple)
    {
      printf("\ntesting for multiple events:\n");
      printf("----------------------------\n");

      /* define events to be measured */
      counter_list[0] = PCL_CYCLES;
#if defined(SPARC)
      counter_list[1] = PCL_INSTR;
#else
      counter_list[1] = PCL_FP_INSTR;
#endif

      /* query for functionality */
      if ((res = PCLquery(counter_list, 2, flags)) != PCL_SUCCESS)
	printf("two events not possible (%s)\n", errornames[-res]);
      else
	{
	  /* start counting */
	  if ((res = PCLstart(counter_list, 2, flags)) != PCL_SUCCESS)
	    {
	      printf("problem with starting two events (%s)\n",
		     errornames[-res]);
	      exit(1);
	    }

	  /* do work */
	  matadd(a, b, c);

	  /* stop counting */
	  if ((res = PCLstop(result, 2)) != PCL_SUCCESS)
	    printf("problem with stopping two events (%s)\n", errornames[-res]);
	  else
	    printf(OFORMAT " cycles with " OFORMAT " "
#if defined(SPARC)
		   "instructions\n",
#else
		   "floating point operations\n",
#endif
		   (OTYPE) result[0], (OTYPE) result[1]);
	}
    }


/*-----------------------------------------------------------*/
  /* test 32-bit overflow */

  if (do_overflow)
    {
      printf("\ntesting 32-bit overflow (may take a while):\n");
      printf("-------------------------------------------\n");

      /* count cycles */
      event = counter_list[0] = PCL_CYCLES;

      /* start counting */
      if ((res = PCLstart(counter_list, 1, flags)) != PCL_SUCCESS)
	{
	  printf("problem with starting event %d (%s)\n",
		 event, errornames[-res]);
	  exit(1);
	}

      /* do work */
      {
	unsigned int i, j, k, ijk;

	/* We try to get 2^32 cycles for sure. */
	ijk = 0;
	for(i = 2048; i > 0; --i)
	  for(j = 2048; j > 0; --j)
	    for(k = 1024; k > 0; --k)
	      ijk += i - j + k;
	
	/* some compilers are smart */
	dummy(ijk);
      }

      /* stop counting */
      if ((res = PCLstop(&result[0], 1)) != PCL_SUCCESS)
	printf("problem with stopping event %d (%s)\n",
	       event, errornames[-res]);
      else
	printf(OFORMAT " cycles executed\n",
	       (OTYPE) result[0]);
    }


/*-----------------------------------------------------------*/

  return 0;
}


/*                                                                            */
/*============================================================================*/
/*                                                                            */

static void
dummy(unsigned int ijk)
{
}


/*                                                                            */
/*============================================================================*/
